#include "UrsArduinoSoftI2c.h"

// Frequency  = 105 kHz, period = 9.5 us at a 16 MHz Arduino Uno

bool UrsArduinoSoftI2cClass::begin(uint8_t sda, uint8_t scl) {
   // In order not to have to consider the current AVR type: get register pointers and bit mask with Arduino functions.
   sdaMask = digitalPinToBitMask(sda);
   sclMask = digitalPinToBitMask(scl);

   uint8_t port = digitalPinToPort(sda);
   if (port == NOT_A_PIN) return false;
   sdaDDR = portModeRegister(port);
   sdaPIN = portInputRegister(port);
   volatile uint8_t* sdaPORT = portOutputRegister(port);

   port = digitalPinToPort(scl);
   if (port == NOT_A_PIN) return false;
   sclDDR = portModeRegister(port);

   sclPIN = portInputRegister(port);

   volatile uint8_t* sclPORT = portOutputRegister(port);


   *sdaPORT &= ~(SoftI2c.sdaMask); // When the pin is switched to OUTPUT, it goes LOW.
   *sclPORT &= ~(SoftI2c.sclMask); // ... at INPUT, the external PULLUP pulls the level to HIGH.

   pinMode(sda, INPUT);
   pinMode(scl, INPUT);
   // Start with INPUT, i.e. both pins on HIGH.
   sdaHigh();

   sclHigh();
   return true;
}

uint8_t UrsArduinoSoftI2cClass::writePacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len, bool sendStop) {
   uint8_t count = 0;

   Start();
   if (!Write((deviceAdr << 1) | 0x00)) { // device address + write signal
      goto Exit; // count is 0
   }
   for (uint8_t i = 0; i < len; i++) {
      if (!Write(pData[i]))
         goto Exit;
      else
         count++;
   }

Exit:
   if (sendStop)
      Stop();
   return count;
}

uint8_t UrsArduinoSoftI2cClass::readPacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len) {
   Start();

   if (!Write((deviceAdr << 1) | 0x01)) { // device address + read signal
      Stop();
      return 0;
   }
   for (uint8_t i = 0; i < len - 1; i++) { // Das letzte Byte wird separat eingelesen
      *pData++ = Read(true); // true: further bytes are requested
   }
   *pData = Read(false); // last Byte

   Stop();
   return len;
}

uint8_t UrsArduinoSoftI2cClass::writeToRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len) {
   uint8_t count = 0;

   Start();
   if (!Write((deviceAdr << 1) | 0x00)) // device address + write signal
      goto Exit; // count is 0
   if (!Write(regIndex))
      goto Exit; // count is 0

   for (uint8_t i = 0; i < len; i++) {
      if (!Write(pData[i]))
         goto Exit;
      else
         count++;
   }

Exit:
   Stop();
   return count;
}

uint8_t UrsArduinoSoftI2cClass::readFromRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len, bool repeatedStart) {
   uint8_t rtc = writePacket(deviceAdr, &regIndex, 1, !repeatedStart); // Set Register index
   if (rtc == 0) {
      Stop();
      return 0;
   }
   return readPacket(deviceAdr, pData, len);
}

void UrsArduinoSoftI2cClass::Start() {
   Serial.println("Start");
   sclHigh();
   sdaLow();
}

void UrsArduinoSoftI2cClass::Stop() {
   Serial.println("Stop");
   sdaLow();
   sclHigh();
   sdaHigh();
}

uint8_t UrsArduinoSoftI2cClass::Write(uint8_t data) {
   for (uint8_t i = 0; i < 8; i++)
   {
      sclLow();

      if (data & 0x80)
         sdaHigh();
      else
         sdaLow();

      sclHigh();
      data = data << 1;
   }

   //The 9th clock (ACK Phase)
  sclLow();

   sdaHigh();

   sclHigh();

   uint8_t ack = !((*sdaPIN & sdaMask) > 0);

   sclLow();

   return ack;
}

uint8_t UrsArduinoSoftI2cClass::Read(bool ack) {
   uint8_t data = 0x00;

   for (uint8_t i = 0; i < 8; i++)
   {
      sclLow();
      sclHigh();
      data = data << 1;
      if (*sdaPIN & sdaMask)
         data |= 0x01;
   }

   sclLow();

   if (ack)
   {
      sdaLow();
   }
   else
   {
      sdaHigh();
   }

   sclHigh();

   sclLow();
   sdaHigh(); // Release SDA

   return data;
}

void UrsArduinoSoftI2cClass::sdaLow() {
   *sdaDDR |= sdaMask;
   _delay_loop_2(delayLoops);
}

void UrsArduinoSoftI2cClass::sdaHigh() {
   *sdaDDR &= ~sdaMask;
   _delay_loop_2(delayLoops);
}

void UrsArduinoSoftI2cClass::sclLow() {
   *sclDDR |= sclMask;
   _delay_loop_2(delayLoops);
}

void UrsArduinoSoftI2cClass::sclHigh() {
   *sclDDR &= ~sclMask;
   _delay_loop_2(delayLoops);
   while ((*sclPIN & sclMask) == 0); // Clock stretching
}


UrsArduinoSoftI2cClass SoftI2c;
